#!/usr/bin/env python
# Logistic Regression

import os
import sys
import logging
import liblinearutil as llu

from plot_roc import *
from compute_perf_stat import *
from random import randrange, seed
from optparse import OptionParser

from load_feature_index import *
from standardize import standardize

g_feature_name_index = []


def do_lr_train_predict(train_y, train_x, test_y, test_x, param):
    model = llu.train(train_y, train_x, param)
    pred_labels, evals, deci_values = llu.predict(test_y, test_x, model)
    labels = model.get_labels()
    posi_deci_values = [labels[0] * val[0] for val in deci_values]
    stat = stat_compute(test_y, pred_labels)
    return model, posi_deci_values, stat


def save_partitions(partitions):
    for i in range(len(partitions)):
        train_index, test_index = partitions[i][4:]
        with open("train_idx_%d" % i, 'w') as f:
            for idx in train_index:
                print >> f, idx

        with open("test_idx_%d" % i, 'w') as f:
            for idx in test_index:
                print >> f, idx


def cv_partition(y, x, nfold):
    # set fixed seed to generage same partition every run
    seed(0)

    ylen = len(y)
    indices = range(ylen)
    partitions = []

    # random permutation by swapping i and j instance
    for i in range(ylen):
        j = randrange(i, ylen)
        x[i], x[j] = x[j], x[i]
        y[i], y[j] = y[j], y[i]
        indices[i], indices[j] = indices[j], indices[i]

    for i in range(nfold):
        begin = i * ylen // nfold
        end = (i + 1) * ylen // nfold
        train_x = x[:begin] + x[end:]
        train_y = y[:begin] + y[end:]
        train_idx = indices[:begin] + indices[end:]
        test_x = x[begin:end]
        test_y = y[begin:end]
        test_idx = indices[begin:end]

        partitions.append(
            (train_y, train_x, test_y, test_x, train_idx, test_idx))

    return partitions


def get_partitions(data_dir, nfold, depth=1, suffix='scaled'):
    partitions = []
    if depth == 1:
        basename = os.path.basename(os.path.realpath(data_dir))
    elif depth == 2:
        parent_dir = os.path.split(os.path.realpath(data_dir))[0]
        parent_base = os.path.basename(parent_dir)
        basename = os.path.basename(os.path.realpath(data_dir))
        basename = "%s-%s" % (parent_base, basename)

    for i in range(nfold):
        train_file = os.path.join(data_dir,
                                  '%s-train-%d-%s' % (basename, i, suffix))
        test_file = os.path.join(data_dir,
                                 '%s-test-%d-%s' % (basename, i, suffix))
        if not os.path.exists(train_file) or not os.path.exists(test_file):
            print train_file, test_file
            print "Please Check data files in %s" % data_dir
            return None
        train_y, train_x = llu.svm_read_problem(train_file)
        test_y, test_x = llu.svm_read_problem(test_file)
        partitions.append((train_y, train_x, test_y, test_x))

    return partitions


def plot_and_save(perf_total, log2c_range, out_dir='.'):
    for i in range(0, 7):
        plot_perf_stat(
            perf_total, None, log2c_range, 'log2c', i, 'single', out_dir)

    plot_perf_stat(perf_total, None, None, None, None, 'roc', out_dir)

    for perf_stat in perf_total:
        perf_stat.save(out_dir)


def save_fpr_tpr(deci, test_y, w, c, iternum):
    fprs, tprs, auc = roc_curve(deci, test_y)
    with open('FPRs-TPRs-W-%d-C-%g-iter-%d.txt' % (w, c, iternum), 'w') as f:
        for i in xrange(len(fprs)):
            print >>f, fprs[i], tprs[i]


def save_temp_result(outdir, weight, c, k, model, train_y, perf):
    global g_feature_name_index

    file_name_suffix = "_W-%d,C-%g,k-%d" % (weight, c, k)
    model_file = os.path.join(outdir, "model%s.txt" % (file_name_suffix))
    deci_values_file = os.path.join(
        outdir, "deci_values%s.txt" % (file_name_suffix))
    actual_label_file = os.path.join(
        outdir, 'actual_label%s.txt' % (file_name_suffix))
    W_file = os.path.join(outdir, "W%s.txt" % (file_name_suffix))

    llu.save_model(model_file, model)

    with open(deci_values_file, 'w') as f:
        for v in perf.deci_value[k]:
            print >>f, v

    with open(actual_label_file, 'w') as f:
        for t in perf.actual_data[k]:
            print >>f, t

#    W = model.get_W(train_y)
#    with open(W_file, 'w') as f:
#        w_sorted = sorted(W.iteritems(),
#                          key=lambda x: abs(x[1]), reverse=True)
#        for index, wi in w_sorted:
#            print >>f, index, wi, g_feature_name_index[index]


def do_cross_validation(partitions, weight, cost,
                        do_standardizing=False, out_dir='.'):
    nfold = len(partitions)

    perf_total = []

    L1LR_params = "-s 6 -c %g -w1 %d -w-1 1" % (cost, weight)
    perf = Performance('L1LR', 'w1-%d,C-%g' %
                       (weight, cost), [], [], [], nfold, L1LR_params)
    logging.info("L1LR %s" % L1LR_params)

    for k in range(nfold):
        print L1LR_params, "iter: %d" % k
        train_y, train_x, test_y, test_x = partitions[k][:4]

        if do_standardizing:
            train_xz, test_xz = standardize(train_x, test_x)
            model, deci, evals = do_lr_train_predict(train_y, train_xz,
                                                     test_y, test_xz,
                                                     L1LR_params)
        else:
            model, deci, evals = do_lr_train_predict(train_y, train_x,
                                                     test_y, test_x,
                                                     L1LR_params)

        perf.actual_data.append(test_y)
        perf.deci_value.append(deci)
        perf.perf_stat.append(PerfStat(evals))

        save_fpr_tpr(deci, test_y, weight, cost, k)
        save_temp_result(
            out_dir, weight, cost, k, model, train_y, perf)
        logging.info("iter %d done" % (k))

    perf.average()
    perf_total.append(perf)

    for i in range(0, 7):
        plot_perf_stat(
            perf_total, None, [x for x in range(nfold)],
            'iter', i, 'single', out_dir)

    plot_perf_stat(perf_total, None, None, None, None, 'roc', out_dir)

    for perf_stat in perf_total:
        perf_stat.save('.')


def main(options, arguments):
    global g_feature_name_index
    weight = options.weight
    cost = options.cost
    g_feature_name_index = load_feature_index(options.feature_list)

    if options.problem_file:
        prob_y, prob_x = llu.svm_read_problem(options.problem_file)
        # cross validation
        partitions = cv_partition(prob_y, prob_x, options.nfold)
    elif options.normalized_dir:
        if options.top:
            # for normalized topw features experiments
            partitions = get_partitions(options.normalized_dir,
                                        options.nfold, 2)
        else:
            partitions = get_partitions(options.normalized_dir,
                                        options.nfold, 1)
    elif options.raw_data_dir:
        partitions = get_partitions(options.raw_data_dir,
                                    options.nfold, depth=1, suffix='raw')

    if not any([options.normalized_dir, options.raw_data_dir]):
        save_partitions(partitions)

    logging.basicConfig(level=logging.INFO,
                        filename='running.log',
                        format='%(asctime)s %(levelname)-6s %(message)s',
                        datefmt='%Y-%m-%d %H:%M:%S',
                        filemode='w')
    do_cross_validation(partitions, weight, cost, options.standardize)
    logging.shutdown()


if __name__ == "__main__":
    option_0 = {'name': ('-i', '--problem-file'),
                'help': 'input file containing training and testing data',
                'nargs': 1}
    option_1 = {'name': ('-c', '--cost'), 'help': 'set svm param c',
                'default': '8', 'type': 'float', 'nargs': 1}
    option_2 = {'name': ('-v', '--nfold'), 'type': 'int',
                'help': 'fold number of cross validation',
                'default': '5', 'nargs': 1}
    option_3 = {'name': ('-w', '--weight'), 'type': 'int',
                'help': 'weight of class +1', 'default': '2', 'nargs': 1}
    option_4 = {'name': ('-o', '--out-dir'), 'default': './',
                'help': 'output directory', 'nargs': 1}
    option_5 = {'name': ('-t', '--compute-tpr'), 'action': 'store_true',
                'default': False,
                'help': 'check detection rate under given fpr'}
    option_6 = {'name': ('-F', '--feature-list'),
                'help': 'feature index file',
                'nargs': 1}
    option_7 = {'name': ('-s', '--standardize'), 'help': 'do standardizing',
                'default': False, 'action': 'store_true'}
    option_8 = {'name': ('-n', '--normalized-dir'),
                'help': 'load normalized data in this dir',
                'nargs': 1}
    option_9 = {'name': ('-r', '--raw-data-dir'),
                'help': 'load raw data in this dir',
                'nargs': 1}
    option_A = {'name': ('-T', '--top'), 'help': 'top w features',
                'default': False, 'action': 'store_true'}

    options = [option_0, option_1, option_2, option_3,
               option_4, option_5, option_6, option_7,
               option_8, option_9, option_A]

    parser = OptionParser()
    for option in options:
        param = option['name']
        del option['name']
        parser.add_option(*param, **option)

    options, arguments = parser.parse_args()
    sys.argv[:] = arguments

    main(options, arguments)
